/************************************************************************
 *                                                                      *
 *      A TABU SEARCH CODE FOR THE VEHICLE ROUTING PROBLEM		*
 *                                                                      *
 *  based on the Experimental Tabu Search code for the nQueens-problem  *
 *			by  Manuel Laguna                  		*
 *                                                                      *
 *  Last revision: January, 2004                          Version: 1.1  *
 *                                                                      *
 *  TSQUEEN.C :   <n>          number of iterations                     *
 *                <tabu_size>  size of the tabu list                    *
 *              							*
 *                                                                      * 
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define numcus 100				/* number of customers		*/
#define MAX(X,Y)    ( (X) > (Y) ? (X) : (Y) )
#define BIG 9999999

struct shop {
	double X;		/* X-position of the shop	*/
	double Y;		/* Y-position of the shop	*/
	int open;		/* opening sequence		*/
	int close;		/* closing sequence		*/
	int supply;		/* supply of the shop		*/
	int demand;		/* demand of the shop		*/
	int visited;		/* shop visited?		*/
};

struct move_info {
	int i;
	int j;
	double value;
};


	int iter;                  		/* iteration counter			*/
	int iter_best;		             	/* iteration where best was found	*/
	int sequence[numcus + 2];		/* sequence of the shops		*/
	int best_sequence[numcus + 2];		/* best sequence of the shops		*/
	int tabu_time[numcus + 1][numcus + 1];	/* matrix with tabu moves		*/

	

	double start_solution;			/* starting solution			*/
	double current_solution;		/* current solution at iter		*/
	double best_solution;			/* best solution until now		*/
	double distance[numcus + 1][numcus + 1];/* distance matric between shops	*/
	

	struct shop customers[numcus + 1];	/* array of shops			*/
		




/************************************************************************
 *                                                                      *
 *                       FUNCTION PROTOTYPES                            *
 *                                                                      *
 ************************************************************************/

void tabu_search(int n, int tabu_size);
void initialization();
void solution();
void best_move(int tabu_size);
double evaluate(int i, int j);
void execute_move(int best_i, int best_j, int tabu_size);
int feasibility(int i, int j);

/***************** Memory Allocation Prototypes *************************/

int *ivector(int nl, int nh);
void free_ivector(int *v, int nl);
int **imatrix(int nrl, int nrh, int ncl, int nch);
void free_imatrix(int **m, int nrl, int nrh, int ncl);
void nrerror(char *error_text);


/************************************************************************
 *                                                                      *
 *                         MAIN FUNCTION                                *
 *                                                                      *
 ************************************************************************/

void main(int argc, char **argv)

{
	int n;                  /* number of iterations                     */
	int tabu_size;          /* size of the short term memory function   */
	
	if (argc != 3) nrerror("TSQUEEN: <n> <tabu_size>");
	n = atoi(argv[1]);
	tabu_size = atoi(argv[2]);
	tabu_search(n, tabu_size);
}

/************************************************************************
 *                                                                      *
 *                           TABU SEARCH                                *
 *                                                                      *
 ************************************************************************/

void tabu_search(int n, int tabu_size)

{

	int i,j;		/* counters		*/
	

	initialization();
	solution();
	start_solution = current_solution;
	
	/*
	for (i = 0; i <= n; i++){
	   best_move(tabu_size);
	   solution();
	}
	*/
	
	
	for (i = 0; i <= numcus + 1; i++){
	printf("%d - ",sequence[i]);
	}
	
	printf("\nde beste oplossing is %lf\n", best_solution);



	
}

/************************************************************************
 *                                                                      *
 *                         INITIALIZATION                               *
 *                                                                      *
 ************************************************************************/
	
void initialization()

{
	int i, j, k, position;				/* counters			*/
	int m, succ;						/* used for reading a file	*/
	int firstclose;					/* shop that closest first	*/
	
	double p, d;					/* used for reading a file	*/
	double X_dist, Y_dist;				/* X-distance and Y distance	*/
	
	FILE *fdata;
		
	firstclose = 999;
	fdata = fopen("data.txt", "r");
	j = 1;
	i = 1;
	
	while (i<= numcus){				/* fill the array of customers with the		*/
		while (j<=numcus*7) {			/* corresponding data from the file "data.txt"	*/
		fscanf(fdata, "%d", &m);
		fscanf(fdata, "%lf", &p);
		customers[i].X = p;
		++j;
		fscanf(fdata, "%lf", &p);
		customers[i].Y = p;
		++j;
		fscanf(fdata, "%d", &m);
		customers[i].open = m;
		++j;
		fscanf(fdata, "%d", &m);
		customers[i].close = m;
		++j;
		fscanf(fdata, "%d", &m);
		customers[i].supply = m;
		++j;
		fscanf(fdata, "%d", &m);
		customers[i].demand = m;
		++j;++i;
		}
		
	}
	
	for (i = 0; i <= numcus; i++){				/* calculating the distances to fill the distance matrix */
	   for (j = 0; j <= numcus; j++){
		X_dist = MAX((customers[i].X - customers[j].X), (customers[j].X - customers[i].X));
		Y_dist = MAX((customers[i].Y - customers[j].Y), (customers[j].Y - customers[i].Y));
		distance[i][j] = (X_dist + Y_dist);
	   }
	}
	
	sequence[0] = sequence[numcus + 1] = 0;			/* start and end point of the route is the warehouse (0)*/
	
	
	j = 0;
	
	d = BIG;
while (j < numcus){	
	for (i = 1; i <= numcus; i++){
	   
	   if ((j != i) & (customers[i].visited == 0)){
		  if (distance[j][i] < d){
		  d = distance[j][i];
		  printf("%lf\n", d);
		  succ = i;
		  printf("%d\n", succ);
		}
		}
		

	
	}
		++j;
		sequence[j] = succ;
		
		customers[sequence[j]].visited = 1;
		d=BIG;
	
}	
	
	best_solution = BIG;
        current_solution = 0;
        iter = 1;
       
}
	


/************************************************************************
 *                                                                      *
 *                         CALCULATE SOLUTION                           *
 *                                                                      *
 ************************************************************************/
	
void solution()

{
  int i;

  current_solution = 0;
  for (i = 0; i <= numcus; i++){
		current_solution = current_solution + distance[sequence[i]][sequence[i+1]];
  }
  
  if (current_solution < best_solution){
  	best_solution = current_solution;
  	iter_best = iter;
  	for (i = 0; i < numcus; i++){
  	  best_sequence[i] = sequence[i];
  	}
  }

	
}


/************************************************************************
 *                                                                      *
 *                           BEST MOVE                                  *
 *                                                                      *
 ************************************************************************/

void best_move(int tabu_size)
{

   
   struct move_info move[((numcus*(numcus-1))/2)];     /* information related to best move      */
     
   
   int i, j, k, best_i, best_j, feasible;
   double best_swap;
   
   
   best_swap = BIG;


   k = -1;
   for (i = 1; i <= numcus-1; i++){
		for (j = i+1; j <= numcus; j++){
			++k;
			move[k].i = i;
			move[k].j = j;
			feasible = feasibility(i, j);
			if (tabu_time[i][j] > iter || feasible == 0){
			   move[k].value = BIG;
			}	
			else {
			   move[k].value = evaluate(i,j);
			}
   }
   }
   
   
   for (i = 0; i <= k; i++){
         if (move[i].value < best_swap){
         best_swap = move[i].value;
         best_i = move[i].i;
         best_j = move[i].j;
      }
      }

 /*  printf("%d %d %lf\n", best_i, best_j, best_swap);*/
   if (best_i>100 || best_j>100) { printf(" Algo paso...\n "); exit(1);}
   execute_move(best_i, best_j, tabu_size);
}


/************************************************************************
 *                                                                      *
 *                       MOVE EVALUATION                                *
 *                                                                      *
 ************************************************************************/

double evaluate(int i, int j)

{
	int k, temp;
	double d;

	d = 0;
		
	if ((i == (j + 1)) || (j == (i + 1))){
	d = d - distance[sequence[i-1]][sequence[i]];
	d = d - distance[sequence[j]][sequence[j+1]];
	d = d + distance[sequence[i-1]][sequence[j]]; 
	d = d + distance[sequence[i]][sequence[j+1]];
	
	}
	else {
	d = d - distance[sequence[i-1]][sequence[i]];
	d = d - distance[sequence[i]][sequence[i+1]];
	d = d - distance[sequence[j-1]][sequence[j]];
	d = d - distance[sequence[j]][sequence[j+1]];
	d = d + distance[sequence[i-1]][sequence[j]];
	d = d + distance[sequence[j]][sequence[i+1]];
	d = d + distance[sequence[j-1]][sequence[i]];
	d = d + distance[sequence[i]][sequence[j+1]];
	}

	return d;
    
}

/************************************************************************
 *                                                                      *
 *                        EXECUTE MOVE                                  *
 *                                                                      *
 ************************************************************************/

void execute_move(int best_i, int best_j, int tabu_size)

{
	int i, temp;
	
	
	temp = sequence[best_j];
	sequence[best_j] = sequence[best_i];
	sequence[best_i] = temp;
	
	tabu_time[best_i][best_j] = iter + tabu_size;
	tabu_time[best_j][best_i] = iter + tabu_size;
	
	/*for (i = 0; i <= numcus + 1; i++){
	   printf("%d ", sequence[i]);
	}*/
	
	iter++;
   
}

/************************************************************************
 *                                                                      *
 *                        CHECK FEASIBILITY                             *
 *                                                                      *
 ************************************************************************/

int feasibility(int i, int j)

{
	int f;
		
	if (customers[sequence[j]].open > i || customers[sequence[j]].close < i ||
            customers[sequence[i]].open > j || customers[sequence[i]].close < j){
	   f = 0;
	}
	else {
	   f = 1;
	}
	
	return f;
        
}

/************************************************************************
 *                                                                      *
 *                  MEMORY ALLOCATION FUNCTIONS                         *
 *                                                                      *
 ************************************************************************/

int *ivector(int nl, int nh)

{
	int *v;

	v = (int *)malloc((unsigned)(nh-nl+1)*sizeof(int));
	if (!v) nrerror("Memory allocation failure in ivector()");
	return v - nl;
}

void free_ivector(int *v, int nl)

{
	free((char*) (v + nl));
}

int **imatrix(int nrl, int nrh, int ncl, int nch)

{
	int i;
	int **m;

	m = (int **)malloc((unsigned)(nrh - nrl +1)*sizeof(int*));
	if (!m) nrerror("allocation failure 1 in imatrix()");
	m -= nrl;
	for(i = nrl; i <= nrh ; i++) {
		m[i]= (int*)malloc((unsigned)(nch - ncl +1)*sizeof(int));
		if (!m[i]) nrerror("allocation failure 2 in imatrix()");
		m[i] -= ncl;
	}
	return m;
}

void free_imatrix(int **m, int nrl, int nrh, int ncl)

{
	int i;

	for (i = nrh; i >= nrl; i--)
		free((char*) (m[i] + ncl));
	free((char *)(m + nrl));
}

void nrerror(char *error_text)

{
	fprintf(stderr,"\n%s\n",error_text);
	fprintf(stderr,"... now exiting to system ...\n");
	exit(1);
}
