/************************************************************************
 *                                                                      *
 *      A TABU SEARCH CODE FOR THE VEHICLE ROUTING PROBLEM		*
 *                                                                      *
 *  based on the Experimental Tabu Search code for the nQueens-problem  *
 *			by  Manuel Laguna                  		*
 *                                                                      *
 *  Last revision: January, 2004                          Version: 1.1  *
 *                                                                      *
 *  TSQUEEN.C :   <n>          number of iterations                     *
 *                <tabu_size>  size of the tabu list                    *
 *              							*
 *                                                                      * 
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define numcus 100			/* number of customers		*/
#define MAX(X,Y)    ( (X) > (Y) ? (X) : (Y) )
#define BIG 9999999
#define max_load 50



struct shop {
	double X;		/* X-position of the shop	*/
	double Y;		/* Y-position of the shop	*/
	int open;		/* opening sequence		*/
	int close;		/* closing sequence		*/
	int supply;		/* supply of the shop		*/
	int demand;		/* demand of the shop		*/
	int visited;		/* shop visited?		*/
};

struct move_info {
	int i;
	int j;
	int h;
	double value;
};


	int iter;                  		/* iteration counter			*/
	int iter_best;		             	/* iteration where best was found	*/
	int sequence[numcus + 2];		/* sequence of the shops		*/
	int best_sequence[numcus + 2];		/* best sequence of the shops		*/
	int tabu_time[numcus + 1][numcus + 1][numcus +  1];	/* matrix with tabu moves		*/
	int tabu_time2[numcus + 1][numcus + 1];	/* matrix with tabu moves		*/
	int option;
	int load[numcus + 1];
	int moved[numcus + 1];
	
	float bla;
	

	double start_solution;			/* starting solution			*/
	double current_solution;		/* current solution at iter		*/
	double best_solution;			/* best solution until now		*/
	double distance[numcus + 1][numcus + 1];/* distance matric between shops	*/
	

	struct shop customers[numcus + 1];	/* array of shops			*/
	

		




/************************************************************************
 *                                                                      *
 *                       FUNCTION PROTOTYPES                            *
 *                                                                      *
 ************************************************************************/

void tabu_search(int n, int tabu_size);
void initialization();
void solution();
void best_move3(int tabu_size, float bla);
double evaluate3(int i, int j, int h);
void execute_move3(int best_i, int best_j, int best_h, int best_option, int tabu_size);
int feasibility3(int i, int j, int h);
void best_move2(int tabu_size);
double evaluate2(int i, int j);
void execute_move2(int best_i, int best_j, int tabu_size);
int feasibility2(int i, int j);

/***************** Memory Allocation Prototypes *************************/

int *ivector(int nl, int nh);
void free_ivector(int *v, int nl);
int **imatrix(int nrl, int nrh, int ncl, int nch);
void free_imatrix(int **m, int nrl, int nrh, int ncl);
void nrerror(char *error_text);


/************************************************************************
 *                                                                      *
 *                         MAIN FUNCTION                                *
 *                                                                      *
 ************************************************************************/

void main(int argc, char **argv)

{
	int n;                  /* number of iterations                     */
	int tabu_size;          /* size of the short term memory function   */
	
	if (argc != 3) nrerror("TSQUEEN: <n> <tabu_size>");
	n = atoi(argv[1]);
	tabu_size = atoi(argv[2]);
	
	tabu_search(n, tabu_size);
}

/************************************************************************
 *                                                                      *
 *                           TABU SEARCH                                *
 *                                                                      *
 ************************************************************************/

void tabu_search(int n, int tabu_size)

{

	int i,j;		/* counters		*/
	
	
	
	
	initialization();
	
	
	solution();
	start_solution = current_solution;
	
	
	for (j = 0; j <= 5; j++){
	
		for (i = 0; i <= 149; i++){
		   best_move2(tabu_size);
		   solution();
		}
		/*for (i = 1; i < numcus; i++){
  	  	sequence[i] = best_sequence[i];
	  	}*/
		for (i = 0; i <= 3; i++){
	  	   best_move3(tabu_size, bla);
   		   solution();
		}
		
	}
	
	
	for (i = 0; i <= numcus + 1; i++){
	printf("%d - ",sequence[i]);
	}
	
	printf("\nde huidige oplossing is %lf\n", current_solution);
	printf("\nde beste oplossing is %lf\n", best_solution);



	
}

/************************************************************************
 *                                                                      *
 *                         INITIALIZATION                               *
 *                                                                      *
 ************************************************************************/
	
void initialization()

{
	int i, j, k, position;				/* counters			*/
	int m;						/* used for reading a file	*/
	int firstclose;					/* shop that closest first	*/
	int new_load;
	
	double p;					/* used for reading a file	*/
	double X_dist, Y_dist;				/* X-distance and Y distance	*/
	
	FILE *fdata;
	
	srand(1);	
	firstclose = 999;
	
	fdata = fopen("data.txt", "r");
	j = 1;
	i = 1;
	
	while (i<= numcus){				/* fill the array of customers with the		*/
		while (j<=numcus*7) {			/* corresponding data from the file "data.txt"	*/
		fscanf(fdata, "%d", &m);
		fscanf(fdata, "%lf", &p);
		customers[i].X = p;
		++j;
		fscanf(fdata, "%lf", &p);
		customers[i].Y = p;
		++j;
		fscanf(fdata, "%d", &m);
		customers[i].open = m;
		++j;
		fscanf(fdata, "%d", &m);
		customers[i].close = m;
		++j;
		fscanf(fdata, "%d", &m);
		customers[i].supply = m;
		++j;
		fscanf(fdata, "%d", &m);
		customers[i].demand = m;
		++j;++i;
		}
		
	}
	
	for (i = 0; i <= numcus; i++){				/* calculating the distances to fill the distance matrix */
	   for (j = 0; j <= numcus; j++){
		X_dist = MAX((customers[i].X - customers[j].X), (customers[j].X - customers[i].X));
		Y_dist = MAX((customers[i].Y - customers[j].Y), (customers[j].Y - customers[i].Y));
		distance[i][j] = (X_dist + Y_dist);
	   }
	}

	sequence[0] = sequence[numcus + 1] = 0;			/* start and end point of the route is the warehouse (0)*/
	load[0] = 30;
	
	for (position = 1; position <= numcus; position++){	/* find a starting solution that is feasible, visiting first the*/
		for (i = 1; i <= numcus; i++){			/* shops that closes first is picked				*/
			new_load = load[sequence[i-1]] + customers[i].supply - customers[i].demand;
			if ((customers[i].open <= position) & (customers[i].visited == 0) & new_load <= max_load){
				if (customers[i].close < firstclose){
				firstclose = customers[i].close;
				k = i;
				}
				
			}
		
		}
		sequence[position] = k;
		customers[k].visited = 1;
		load[sequence[position]] = load[sequence[position-1]] + customers[k].supply - customers[k].demand;
		
		firstclose = 999;
		
	}
	
	best_solution = BIG;
        current_solution = 0;
        iter = 1;
}
	


	


/************************************************************************
 *                                                                      *
 *                         CALCULATE SOLUTION                           *
 *                                                                      *
 ************************************************************************/
	
void solution()

{
  int i;

  current_solution = 0;
  for (i = 0; i <= numcus; i++){
		current_solution = current_solution + distance[sequence[i]][sequence[i+1]];
  }
  
  if (current_solution < best_solution){
  	best_solution = current_solution;
  	iter_best = iter;
  	for (i = 0; i < numcus; i++){
  	  best_sequence[i] = sequence[i];
  	}
  }

	
}

/************************************************************************
 *                                                                      *
 *                           BEST MOVE                                  *
 *                                                                      *
 ************************************************************************/

void best_move2(int tabu_size)
{

   
   struct move_info move[((numcus*(numcus-1))/2)];     /* information related to best move      */
     
   
   int i, j, k, best_i, best_j, feasible;
   double best_swap;
   
   
   best_swap = BIG;


   k = -1;
   for (i = 1; i <= numcus-1; i++){
		for (j = i+1; j <= numcus; j++){
			++k;
			move[k].i = i;
			move[k].j = j;
			feasible = feasibility2(i, j);
			if (tabu_time2[i][j] > iter || feasible == 0){
			   move[k].value = BIG;
			}	
			else {
			   move[k].value = evaluate2(i,j);
			}
   }
   }
   
   
   for (i = 0; i <= k; i++){
         if (move[i].value < best_swap){
         best_swap = move[i].value;
         best_i = move[i].i;
         best_j = move[i].j;
      }
      }

 /*  printf("%d %d %lf\n", best_i, best_j, best_swap);*/
   if (best_i>100 || best_j>100) { printf(" Algo paso...\n "); exit(1);}
   execute_move2(best_i, best_j, tabu_size);
}


/************************************************************************
 *                                                                      *
 *                       MOVE EVALUATION                                *
 *                                                                      *
 ************************************************************************/

double evaluate2(int i, int j)

{
	int k, temp;
	double d;

	d = 0;
		
	if ((i == (j + 1)) || (j == (i + 1))){
	d = d - distance[sequence[i-1]][sequence[i]];
	d = d - distance[sequence[j]][sequence[j+1]];
	d = d + distance[sequence[i-1]][sequence[j]]; 
	d = d + distance[sequence[i]][sequence[j+1]];
	
	}
	else {
	d = d - distance[sequence[i-1]][sequence[i]];
	d = d - distance[sequence[i]][sequence[i+1]];
	d = d - distance[sequence[j-1]][sequence[j]];
	d = d - distance[sequence[j]][sequence[j+1]];
	d = d + distance[sequence[i-1]][sequence[j]];
	d = d + distance[sequence[j]][sequence[i+1]];
	d = d + distance[sequence[j-1]][sequence[i]];
	d = d + distance[sequence[i]][sequence[j+1]];
	}

	return d;
    
}

/************************************************************************
 *                                                                      *
 *                        EXECUTE MOVE                                  *
 *                                                                      *
 ************************************************************************/

void execute_move2(int best_i, int best_j, int tabu_size)

{
	int i, temp;
	
	
	temp = sequence[best_j];
	sequence[best_j] = sequence[best_i];
	sequence[best_i] = temp;
	
	load[sequence[best_i]] = (load[sequence[best_i-1]] + customers[best_i].supply - customers[best_j].demand);
	load[sequence[best_j]] = (load[sequence[best_j-1]] + customers[best_j].supply - customers[best_j].demand);
	
	tabu_time2[best_i][best_j] = iter + tabu_size;
	tabu_time2[best_j][best_i] = iter + tabu_size;
	
	moved[best_i] = moved[best_i] + 1;
	moved[best_j] = moved[best_j] + 1;
	
	/*for (i = 0; i <= numcus + 1; i++){
	   printf("%d ", sequence[i]);
	}*/
	
	iter++;
}

/************************************************************************
 *                                                                      *
 *                        CHECK FEASIBILITY                             *
 *                                                                      *
 ************************************************************************/

int feasibility2(int i, int j)

{
	int f;
	
		
	if (customers[sequence[j]].open > i || customers[sequence[j]].close < i ||
            customers[sequence[i]].open > j || customers[sequence[i]].close < j ||
            load[sequence[i-1]] + customers[j].supply - customers[j].demand > max_load ||
            load[sequence[j-1]] + customers[i].supply - customers[i].demand > max_load) {
            f = 0;

	}
	
	else {
	   f = 1;

	}
	
	return f;
        
}



/************************************************************************
 *                                                                      *
 *                           BEST MOVE                                  *
 *                                                                      *
 ************************************************************************/

void best_move3(int tabu_size, float bla)
{

struct move_info move[99999];     /* information related to best move      */
     
   
   int i, j, h, k, l,best_i, best_j, best_h, feasible, best_option;
 
   double best_swap, move_value, test;
   
   
   
   best_swap = BIG;

 
   k = -1;
   l = -1;
   for (i = 1; i <= numcus-2; i++){
	for (j = i+1; j <= numcus-1; j++){
	    for (h = j+1; h <= numcus; h++){	
		
		
			++l;
			feasible = feasibility3(i, j, h);
			move_value = evaluate3(i,j,h);
			if ((feasible == 1) & (tabu_time[i][j][h] < iter) & (move_value < best_swap)){
			  
			best_swap = move_value;
         		best_i = i;
         		best_j = j;
         		best_h = h;
         		best_option = option;
         		test = evaluate3(i, j, h);
         		
         		}
   		
   }
   }
   }
   
   /*printf(" \n\n beste swap levert op i %d j %d h %d swap %lf \n\n", best_i, best_j, best_h, best_swap);*/
   
   if (best_i>100 || best_j>100) { printf(" Algo paso...\n "); exit(1);}
   execute_move3(best_i, best_j, best_h, best_option, tabu_size);
}


/************************************************************************
 *                                                                      *
 *                       MOVE EVALUATION                                *
 *                                                                      *
 ************************************************************************/

double evaluate3(int i, int j, int h)

{
	
	
	
	int k, temp;
	double a, b, c, d, e, f;
	
	
	
	b = 0;
	c = 0;
	
	d = 0;
	

	temp = sequence[i];
	sequence[i] = sequence[j];
	sequence[j] = sequence[h];
	sequence[h] = temp;
	


	
  	for (k = 0; k <= numcus; k++){
		b = b + distance[sequence[k]][sequence[k+1]];
	}
	
	
		
	temp = sequence[j];
	sequence[j] = sequence[h];
	sequence[h] = sequence[i];
	sequence[i] = temp;
	
	
	
	for (k = 0; k <= numcus; k++){
		c = c + distance[sequence[k]][sequence[k+1]];
	}
	
	temp = sequence[h];
	sequence[h] = sequence[i];
	sequence[i] = sequence[j];
	sequence[j] = temp;
	

		
		
	if (b < c){
	  f = b;
	  option = 1;
	}
	else {
	  f = c;
	  option = 2;
	}
	
	e = f - current_solution;
	

	
	return e;
    
}

/************************************************************************
 *                                                                      *
 *                        EXECUTE MOVE                                  *
 *                                                                      *
 ************************************************************************/

void execute_move3(int best_i, int best_j, int best_h, int best_option, int tabu_size)

{
	int i, temp;
	
	if (best_option == 1){
		temp = sequence[best_i];
		sequence[best_i] = sequence[best_j];
		sequence[best_j] = sequence[best_h];
		sequence[best_h] = temp;
		
		
		
	}
	if (best_option == 2){
		temp = sequence[best_i];
		sequence[best_i] = sequence[best_h];
		sequence[best_h] = sequence[best_j];
		sequence[best_j] = temp;
	}
	
	
	load[sequence[best_i]] = (load[sequence[best_i-1]] + customers[best_i].supply - customers[best_j].demand);
	load[sequence[best_j]] = (load[sequence[best_j-1]] + customers[best_j].supply - customers[best_j].demand);
	load[sequence[best_h]] = (load[sequence[best_h-1]] + customers[best_h].supply - customers[best_h].demand);

	
	
	tabu_time[best_i][best_j][best_h] = iter + tabu_size;
	tabu_time[best_i][best_h][best_j] = iter + tabu_size;
	tabu_time[best_j][best_i][best_h] = iter + tabu_size;
	tabu_time[best_j][best_h][best_i] = iter + tabu_size;
	tabu_time[best_h][best_i][best_j] = iter + tabu_size;
	tabu_time[best_h][best_j][best_i] = iter + tabu_size;	
	
	
	iter++;
   
}

/************************************************************************
 *                                                                      *
 *                        CHECK FEASIBILITY                             *
 *                                                                      *
 ************************************************************************/

int feasibility3(int i, int j, int h)

{
	int f;
		
	if (option == 1){
		if (customers[sequence[j]].open > i || customers[sequence[j]].close < i ||
      		    customers[sequence[i]].open > j || customers[sequence[i]].close < j ||
		    customers[sequence[h]].open > h || customers[sequence[h]].close < h ||
		    load[sequence[i-1]] + customers[j].supply - customers[j].demand > max_load ||
		    load[sequence[j-1]] + customers[h].supply - customers[h].demand > max_load ||
		    load[sequence[h-1]] + customers[i].supply - customers[i].demand > max_load)
		{
		f = 0;
		}
		else {
		   f = 1;
		}
	}
	
	if (option == 2){
		if (customers[sequence[j]].open > sequence[i] || customers[sequence[j]].close < sequence[i] ||
        	    customers[sequence[i]].open > sequence[j] || customers[sequence[i]].close < sequence[j] ||
	            customers[sequence[h]].open > sequence[h] || customers[sequence[h]].close < sequence[h] ||
		    load[sequence[i-1]] + customers[h].supply - customers[h].demand > max_load ||
		    load[sequence[j-1]] + customers[i].supply - customers[i].demand > max_load ||
		    load[sequence[h-1]] + customers[j].supply - customers[j].demand > max_load)
	            {
		   f = 0;
		}
		else {
		   f = 1;
		}
	}
	
	
	
	
	
	return f;
        


}

/************************************************************************
 *                                                                      *
 *                  MEMORY ALLOCATION FUNCTIONS                         *
 *                                                                      *
 ************************************************************************/

int *ivector(int nl, int nh)

{
	int *v;

	v = (int *)malloc((unsigned)(nh-nl+1)*sizeof(int));
	if (!v) nrerror("Memory allocation failure in ivector()");
	return v - nl;
}

void free_ivector(int *v, int nl)

{
	free((char*) (v + nl));
}

int **imatrix(int nrl, int nrh, int ncl, int nch)

{
	int i;
	int **m;

	m = (int **)malloc((unsigned)(nrh - nrl +1)*sizeof(int*));
	if (!m) nrerror("allocation failure 1 in imatrix()");
	m -= nrl;
	for(i = nrl; i <= nrh ; i++) {
		m[i]= (int*)malloc((unsigned)(nch - ncl +1)*sizeof(int));
		if (!m[i]) nrerror("allocation failure 2 in imatrix()");
		m[i] -= ncl;
	}
	return m;
}

void free_imatrix(int **m, int nrl, int nrh, int ncl)

{
	int i;

	for (i = nrh; i >= nrl; i--)
		free((char*) (m[i] + ncl));
	free((char *)(m + nrl));
}

void nrerror(char *error_text)

{
	fprintf(stderr,"\n%s\n",error_text);
	fprintf(stderr,"... now exiting to system ...\n");
	exit(1);
}
