/************************************************************************
 *                                                                      *
 *      A TABU SEARCH CODE FOR THE TRAVELING SALESMAN PROBLEM		*
 *			WITH TIME WINDOWS				*
 *                                                                      *
 *  									*
 *					                  		*
 *                                                                      *
 *  Last revision: 25 February, 2004                          Version: 1.0 *
 *                                                                      *
 *  TSQUEEN.C :   <n>          max # non_improve per algorithm          *
 *                <tabu_size>  size of the tabu list                    *
 *              							*
 *                                                                      * 
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#define numcus 100				/* number of customers		*/
#define MAX(X,Y)    ( (X) > (Y) ? (X) : (Y) )
#define BIG 9999999



struct shop {
	double X;		/* X-position of the customer	*/
	double Y;		/* Y-position of the customer	*/
	int open;		/* opening sequence		*/
	int close;		/* closing sequence		*/
	int supply;		/* supply of the customer	*/
	int demand;		/* demand of the customer	*/
	int visited;		/* customer visited?		*/
};

struct move_info {
	int i;
	int j;
	double value;
};


	int iter;                  		/* iteration counter					*/
	int iter_zwaar;
	int iter_best;		             	/* iteration where best was found			*/
	int sequence[numcus + 2];		/* sequence of the shops				*/
	int test_sequence[numcus+2];	 	/* sequence used for testing swaps in reverse_chain()	*/
	int best_sequence[numcus + 2];		/* best sequence of the shops				*/
	int tabu_time[numcus + 1][numcus + 1];	/* matrix with tabu moves				*/
	int no_improve;				/* number of moves that the objective didn't improve	*/
	int moved[numcus + 1];
	

	double start_solution;			/* starting solution			*/
	double current_solution;		/* current solution at iter		*/
	double best_solution;			/* best solution until now		*/
	double distance[numcus + 1][numcus + 1];/* distance matric between shops	*/
	
	clock_t save_clock;
	
	
	

	struct shop customers[numcus + 1];	/* array of customers			*/
		




/************************************************************************
 *                                                                      *
 *                       FUNCTION PROTOTYPES                            *
 *                                                                      *
 ************************************************************************/

void tabu_search(int n, int tabu_size, int gap);
void initialization();
void solution(int tabu_size);
void best_move(int tabu_size);
double evaluate(int i, int j);
void execute_move(int best_i, int best_j, int tabu_size);
int feasibility(int i, int j);
void reverse_chain(int tabu_size, int gap);
double solution2();
int feas();
void back_to_best();
void nrerror(char *error_text);


/************************************************************************
 *                                                                      *
 *                         MAIN FUNCTION                                *
 *                                                                      *
 ************************************************************************/

void main(int argc, char **argv)

{
	int n;                  /* number of iterations                     */
	int tabu_size;          /* size of the short term memory function   */
	int gap;		/* minimum length of chain that is reversed	*/
	
	if (argc != 4) nrerror("TSQUEEN: <n> <tabu_size> <gap>");
	n = atoi(argv[1]);
	tabu_size = atoi(argv[2]);
	gap = atoi(argv[3]);

	
	tabu_search(n, tabu_size, gap);
}

/************************************************************************
 *                                                                      *
 *                           TABU SEARCH                                *
 *                                                                      *
 ************************************************************************/

void tabu_search(int n, int tabu_size, int gap)

{
		
	int i,j;		/* counters					*/

	FILE *ofp;
	
	double user_time;
	double clocks_per_second = (double)CLOCKS_PER_SEC;

	
	
	
	

	
	initialization();
	solution(tabu_size);

	
	
	
	
	start_solution = current_solution;

	ofp = fopen("results.txt", "w");
	

	fprintf(ofp, "These results contain the following values of the variables: n = %d, tabu_size = %d and gap = %d\n\n", n, tabu_size, gap);
	
	for (j  = 0; j <= 0; j++){		
		
		
		while (no_improve < n){
			best_move(tabu_size);
			solution(tabu_size);
			fprintf(ofp, "%0.1lf\n", current_solution);
		}
		
		
		back_to_best();
		no_improve = 0;
		
		
		
		
		
		while (no_improve < n){
			reverse_chain(tabu_size, gap);
			solution(tabu_size);
			fprintf(ofp, "%0.1lf\n", current_solution);
		}
		back_to_best();
		no_improve = 0;
		
	}
	
	
	user_time = save_clock / clocks_per_second;
	
	fprintf(ofp, "number of iterations = %d\n", iter-1);
	fprintf(ofp, "number of iterations until best solution = %d\n", iter_best);
	fprintf(ofp, "Time until best solution = %.1lf\n", user_time);
	
	
	/*save_clock = clock();*/
	
	printf("uitkomst time = %.1lf\n", user_time);
	
	fprintf(ofp, "end of program\n\n");
	
	
	 
	printf("\nEnd Solution\n");
	for (i = 0; i <= numcus + 1; i++){
	printf("%d - ",best_sequence[i]);
	}
	
	fprintf(ofp, "\nEnd Solution\n");
	for (i = 0; i <= numcus + 1; i++){
	fprintf(ofp,"%d - ",best_sequence[i]);
	}
	
	fprintf(ofp,"\nbest solution =  %0.1lf\n", best_solution);
	printf("\nde beste oplossing is %0.1lf\n", best_solution);
		printf("\nhet aantal iteraties is %d\n", iter);
		printf("\nhet aantal zware iteraties is %d\n", iter_zwaar);
		printf("\nhet aantal iteraties tot de beste is %d\n", iter_best);
		
		
	
	printf("%d\n", no_improve);
		
	
	

	
}

/************************************************************************
 *                                                                      *
 *                         INITIALIZATION                               *
 *                                                                      *
 ************************************************************************/
	
void initialization()

{
	int i, j, k, position;				/* counters			*/
	int m;						/* used for reading a file	*/
	int firstclose;					/* shop that closest first	*/

	
	double p;					/* used for reading a file	*/
	double X_dist, Y_dist;				/* X-distance and Y distance	*/
	
	FILE *fdata;
		
	firstclose = 999;
	fdata = fopen("data.txt", "r");
	j = 1;
	i = 1;
	
	while (i<= numcus){				/* fill the array of customers with the		*/
		while (j<=numcus*7) {			/* corresponding data from the file "data.txt"	*/
		fscanf(fdata, "%d", &m);
		fscanf(fdata, "%lf", &p);
		customers[i].X = p;
		++j;
		fscanf(fdata, "%lf", &p);
		customers[i].Y = p;
		++j;
		fscanf(fdata, "%d", &m);
		customers[i].open = m;
		++j;
		fscanf(fdata, "%d", &m);
		customers[i].close = m;
		++j;
		fscanf(fdata, "%d", &m);
		customers[i].supply = m;
		++j;
		fscanf(fdata, "%d", &m);
		customers[i].demand = m;
		++j;++i;
		}
	}
	
	
	
	for (i = 0; i < numcus; i++){
		moved[i] = 0;
	}
	
	
	for (i = 0; i <= numcus; i++){				/* calculating the distances to fill the distance matrix */
		for (j = 0; j <= numcus; j++){
			X_dist = MAX((customers[i].X - customers[j].X), (customers[j].X - customers[i].X));
			Y_dist = MAX((customers[i].Y - customers[j].Y), (customers[j].Y - customers[i].Y));
			distance[i][j] = X_dist + Y_dist;
	  	 }
	}

	sequence[0] = sequence[numcus + 1] = 0;			/* start and end point of the route is the warehouse (0)*/

	
	for (position = 1; position <= numcus; position++){	/* find a starting solution that is feasible, visiting first the*/
		for (i = 1; i <= numcus; i++){			/* shops that closes first is picked				*/
			if ((customers[i].open <= position) & (customers[i].visited == 0)){
				if (customers[i].close < firstclose){
				firstclose = customers[i].close;
				k = i;
				}
				
			}
		
		}
		sequence[position] = k;
		customers[k].visited = 1;
		firstclose = 999;
	}
	
	best_solution = BIG;
        current_solution = 0;
        iter = 1;
        iter_zwaar = 0;
}
	


/************************************************************************
 *                                                                      *
 *                         CALCULATE SOLUTION                           *
 *                                                                      *
 ************************************************************************/
	
void solution(int tabu_size)

{
	int i;
	
	

	current_solution = 0;
	for (i = 0; i <= numcus; i++){
		current_solution = current_solution + distance[sequence[i]][sequence[i+1]];
	}
  
	if (current_solution < best_solution){
		best_solution = current_solution;
		no_improve = 0;
		iter_best = iter;
  				
		for (i = 0; i < numcus + 1; i++){
  		  best_sequence[i] = sequence[i];
  		}
  		save_clock = clock();

  		}
 	else{
 		++no_improve;
 		
 	}
 	
	
}

/************************************************************************
 *                                                                      *
 *                         BACK TO BEST SOLUTION                        *
 *                                                                      *
 ************************************************************************/
	
void back_to_best()

{
	int i, j;
  
  	for (i = 0; i < numcus + 1; i++){
		sequence[i] = best_sequence[i];
  	}
  	
  	for (i = 0; i <= numcus; i++){
  		  for (j = 0; j <= numcus; j++){
  		  	tabu_time[i][j] = 0;
  		  }
  	}
  

	
}


/************************************************************************
 *                                                                      *
 *               EVALUATE SOLUTION FOR REVERSE CHAIN  		        *
 *                                                                      *
 ************************************************************************/
	
double solution2()

{
	int i;
	double test;

	test = 0;
	for (i = 0; i <= numcus; i++){
		test = test + distance[test_sequence[i]][test_sequence[i+1]];
	}
  
  
  	return test;
	
}


/************************************************************************
 *                                                                      *
 *                           BEST MOVE                                  *
 *                                                                      *
 ************************************************************************/

void best_move(int tabu_size)
{

   
	struct move_info move[((numcus*(numcus-1))/2)];     /* information related to best move      */
     
   
	int i, j, k, best_i, best_j, feasible;
	double best_swap;
   
   
	best_swap = BIG;


	k = -1;
	for (i = 1; i <= numcus-1; i++){
		for (j = i+1; j <= numcus; j++){
			++k;
			move[k].i = i;
			move[k].j = j;
			feasible = feasibility(i, j);
			if (tabu_time[i][j] > iter || feasible == 0){
				move[k].value = BIG;
			}	
			else {
				move[k].value = evaluate(i,j);
			}
		}
	}
   
   
	for (i = 0; i <= k; i++){
		if (move[i].value < best_swap){
			best_swap = move[i].value;
			best_i = move[i].i;
			best_j = move[i].j;
      		}
	}

 /*  printf("%d %d %lf\n", best_i, best_j, best_swap);*/
   if (best_i>100 || best_j>100) { printf(" Algo paso...\n "); exit(1);}
   execute_move(best_i, best_j, tabu_size);
}



/************************************************************************
 *                                                                      *
 *                           REVERSE CHAIN                             *
 *                                                                      *
 ************************************************************************/

void reverse_chain(int tabu_size, int gap)
{

   
	int i, j, k,d, l,best_i, best_j, time_ok, tabu_ok, feasible;
 	int loop_sequence[numcus+2];


  
	double best_swap, loop_solution, test_solution;

   
   
	best_swap = BIG;
	loop_solution = BIG; 
	best_i = 0;
	best_j = 0;
 
	for (i = 1; i <= numcus-gap; i++){
		for (j = i+gap; j <= numcus; j++){
			for (k = 0; k < i; k++){ 
				test_sequence[k] = sequence[k];
		   	}
		   	d = j; 					/*j-1*/
		   	for (k = i; k <= j; k++){			/*k < j*/
		   	test_sequence[k] = sequence[d];
		   	--d;
			}
   	
   
			for (k = j+1; k <= numcus + 1; k++){		/*k = j*/
				test_sequence[k] = sequence[k];
			}
			   
			   
			time_ok = feas();
		  
			if (time_ok == 1){
				tabu_ok = 0;
			   	if (tabu_time[i][j] <= iter){
					tabu_ok = 1;
				}
				test_solution = solution2();
				if (tabu_ok == 1 || test_solution < best_solution){
					if (test_solution < loop_solution){
						loop_solution = test_solution;
				   		best_i = i;
				   		best_j = j;
				   		for (k = 0; k <= numcus + 1; k++){
				   	 		loop_sequence[k] = test_sequence[k];
				   		}
				   	}
				}
				
			}
		   
		}
	}
	
	tabu_time[best_i][best_j] = iter + tabu_size;
	tabu_time[best_j][best_i] = iter + tabu_size;
	

	
	++moved[best_i];
	++moved[best_j];
	
	
	for (i = 0; i <= numcus + 1; i++){
		sequence[i] = loop_sequence[i];
	}
 	++iter;
 	++iter_zwaar;



}

/************************************************************************
 *                                                                      *
 *                       FEASIBILITY CHECK FOR REVERSE CHAIN           *
 *                                                                      *
 ************************************************************************/

int feas()

{
	int feasible, i;
 
	feasible = 1;
		   
	for (i = 1; i <= numcus; i++){
		if ((customers[test_sequence[i]].open > i) || (customers[test_sequence[i]].close < i)){
			feasible = 0;
		}
	}
	return feasible;
}





/************************************************************************
 *                                                                      *
 *                       MOVE EVALUATION                                *
 *                                                                      *
 ************************************************************************/

double evaluate(int i, int j)

{
	int k, temp;
	double d;

	d = 0;
		
	if ((i == (j + 1)) || (j == (i + 1))){
	d = d - distance[sequence[i-1]][sequence[i]];
	d = d - distance[sequence[j]][sequence[j+1]];
	d = d + distance[sequence[i-1]][sequence[j]]; 
	d = d + distance[sequence[i]][sequence[j+1]];
	
	}
	else {
	d = d - distance[sequence[i-1]][sequence[i]];
	d = d - distance[sequence[i]][sequence[i+1]];
	d = d - distance[sequence[j-1]][sequence[j]];
	d = d - distance[sequence[j]][sequence[j+1]];
	d = d + distance[sequence[i-1]][sequence[j]];
	d = d + distance[sequence[j]][sequence[i+1]];
	d = d + distance[sequence[j-1]][sequence[i]];
	d = d + distance[sequence[i]][sequence[j+1]];
	}

	return d;
    
}

/************************************************************************
 *                                                                      *
 *                        EXECUTE MOVE                                  *
 *                                                                      *
 ************************************************************************/

void execute_move(int best_i, int best_j, int tabu_size)

{
	int i, temp;
	
	
	temp = sequence[best_j];
	sequence[best_j] = sequence[best_i];
	sequence[best_i] = temp;
	
	
	tabu_time[best_i][best_j] = iter + tabu_size;
	tabu_time[best_j][best_i] = iter + tabu_size;
	
	++moved[best_i];
	++moved[best_j];
	
	/*for (i = 0; i <= numcus + 1; i++){
	   printf("%d ", sequence[i]);
	}*/
	
	iter++;
}

/************************************************************************
 *                                                                      *
 *                        CHECK FEASIBILITY                             *
 *                                                                      *
 ************************************************************************/

int feasibility(int i, int j)

{
	int f;
	
		
	if (customers[sequence[j]].open > i || customers[sequence[j]].close < i ||
            customers[sequence[i]].open > j || customers[sequence[i]].close < j) {
		f = 0;
	}
	
	else {
		f = 1;
	}
	
	return f;
        
}
/************************************************************************
 *                                                                      *
 *                        ERROR REPORT		                        *
 *                                                                      *
 ************************************************************************/
void nrerror(char *error_text)

{
	fprintf(stderr,"\n%s\n",error_text);
	fprintf(stderr,"... now exiting to system ...\n");
	exit(1);
}

