/************************************************************************
 *                                                                      *
 *      A TABU SEARCH CODE FOR THE VEHICLE ROUTING PROBLEM		*
 *                                                                      *
 *  based on the Experimental Tabu Search code for the nQueens-problem  *
 *			by  Manuel Laguna                  		*
 *                                                                      *
 *  Last revision: January, 2004                          Version: 1.1  *
 *                                                                      *
 *  TSQUEEN.C :   <n>          number of iterations                     *
 *                <tabu_size>  size of the tabu list                    *
 *              							*
 *                                                                      * 
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define numcus 100				/* number of customers		*/
#define MAX(X,Y)    ( (X) > (Y) ? (X) : (Y) )
#define BIG 9999999



struct shop {
	double X;		/* X-position of the shop	*/
	double Y;		/* Y-position of the shop	*/
	int open;		/* opening sequence		*/
	int close;		/* closing sequence		*/
	int supply;		/* supply of the shop		*/
	int demand;		/* demand of the shop		*/
	int visited;		/* shop visited?		*/
};

struct move_info {
	int i;
	int j;
	double value;
};


	int iter;                  		/* iteration counter			*/
	int iter_best;		             	/* iteration where best was found	*/
	int sequence[numcus + 2];		/* sequence of the shops		*/
	int test_sequence[numcus+2];	 
	int best_sequence[numcus + 2];		/* best sequence of the shops		*/
	int tabu_time[numcus + 1];	/* matrix with tabu moves		*/


	double start_solution;			/* starting solution			*/
	double current_solution;		/* current solution at iter		*/
	double best_solution;			/* best solution until now		*/
	double distance[numcus + 1][numcus + 1];/* distance matric between shops	*/
	

	struct shop customers[numcus + 1];	/* array of shops			*/
		




/************************************************************************
 *                                                                      *
 *                       FUNCTION PROTOTYPES                            *
 *                                                                      *
 ************************************************************************/

void tabu_search(int n, int tabu_size);
void initialization();
void solution();
void best_move(int tabu_size);
double evaluate(int i, int j);
void execute_move(int best_i, int best_j, int tabu_size);
int feasibility(int i, int j);
void reverse_chain(int tabu_size);
double solution2();
int feas();
void back_to_best();

void nrerror(char *error_text);


/************************************************************************
 *                                                                      *
 *                         MAIN FUNCTION                                *
 *                                                                      *
 ************************************************************************/

void main(int argc, char **argv)

{
	int n;                  /* number of iterations                     */
	int tabu_size;          /* size of the short term memory function   */
	
	if (argc != 3) nrerror("TSQUEEN: <n> <tabu_size>");
	n = atoi(argv[1]);
	tabu_size = atoi(argv[2]);
	
	tabu_search(n, tabu_size);
}

/************************************************************************
 *                                                                      *
 *                           TABU SEARCH                                *
 *                                                                      *
 ************************************************************************/

void tabu_search(int n, int tabu_size)

{

	int i,j;		/* counters		*/
	
	
	
	
	initialization();
	solution();
	start_solution = current_solution;

	
	
	for (i = 0; i <= n; i++){
		best_move(tabu_size);
		solution();
		printf("%lf\n", current_solution);
	}
	
	printf("\nZo zijn we begonnen\n");
	
	for (i = 0; i <= numcus + 1; i++){
		printf("%d - ", sequence[i]);
	}

	
	back_to_best();
	
	for (i = 0; i <= 45; i++){
		reverse_chain(tabu_size);
		solution();
		printf("%lf\n", current_solution);
	}
	
	
	back_to_best();
	
	for (i = 0; i <= n; i++){
		best_move(tabu_size);
		solution();
		printf("%lf\n", current_solution);
	}
	
	back_to_best();
	
	for (i = 0; i <= 30; i++){
		reverse_chain(tabu_size);
		solution();
		printf("%lf\n", current_solution);
	}
	
	 
		printf("\nZo zijn we geeindigd\n");
	for (i = 0; i <= numcus + 1; i++){
	printf("%d - ",best_sequence[i]);
	}

	
	printf("\nde beste oplossing is %lf\n", best_solution);
		printf("\nhet aantal iteraties is %d\n", iter);
		
		
	best_solution = 0;
	for (i = 0; i <= numcus; i++){
		best_solution = best_solution + distance[best_sequence[i]][best_sequence[i+1]];
	}
	printf("%lf\n", best_solution);




	
}

/************************************************************************
 *                                                                      *
 *                         INITIALIZATION                               *
 *                                                                      *
 ************************************************************************/
	
void initialization()

{
	int i, j, k, position;				/* counters			*/
	int m;						/* used for reading a file	*/
	int firstclose;					/* shop that closest first	*/

	
	double p;					/* used for reading a file	*/
	double X_dist, Y_dist;				/* X-distance and Y distance	*/
	
	FILE *fdata;
		
	firstclose = 999;
	fdata = fopen("data.txt", "r");
	j = 1;
	i = 1;
	
	while (i<= numcus){				/* fill the array of customers with the		*/
		while (j<=numcus*7) {			/* corresponding data from the file "data.txt"	*/
		fscanf(fdata, "%d", &m);
		fscanf(fdata, "%lf", &p);
		customers[i].X = p;
		++j;
		fscanf(fdata, "%lf", &p);
		customers[i].Y = p;
		++j;
		fscanf(fdata, "%d", &m);
		customers[i].open = m;
		++j;
		fscanf(fdata, "%d", &m);
		customers[i].close = m;
		++j;
		fscanf(fdata, "%d", &m);
		customers[i].supply = m;
		++j;
		fscanf(fdata, "%d", &m);
		customers[i].demand = m;
		++j;++i;
		}
		
	}
	
	for (i = 0; i <= numcus; i++){				/* calculating the distances to fill the distance matrix */
		for (j = 0; j <= numcus; j++){
			X_dist = MAX((customers[i].X - customers[j].X), (customers[j].X - customers[i].X));
			Y_dist = MAX((customers[i].Y - customers[j].Y), (customers[j].Y - customers[i].Y));
			distance[i][j] = (X_dist + Y_dist);
	  	 }
	}

	sequence[0] = sequence[numcus + 1] = 0;			/* start and end point of the route is the warehouse (0)*/

	
	for (position = 1; position <= numcus; position++){	/* find a starting solution that is feasible, visiting first the*/
		for (i = 1; i <= numcus; i++){			/* shops that closes first is picked				*/
			if ((customers[i].open <= position) & (customers[i].visited == 0)){
				if (customers[i].close < firstclose){
				firstclose = customers[i].close;
				k = i;
				}
				
			}
		
		}
		sequence[position] = k;
		customers[k].visited = 1;
		firstclose = 999;
	}
	
	best_solution = BIG;
        current_solution = 0;
        iter = 1;
}
	


/************************************************************************
 *                                                                      *
 *                         CALCULATE SOLUTION                           *
 *                                                                      *
 ************************************************************************/
	
void solution()

{
	int i;

	current_solution = 0;
	for (i = 0; i <= numcus; i++){
		current_solution = current_solution + distance[sequence[i]][sequence[i+1]];
	}
  
	if (current_solution < best_solution){
		best_solution = current_solution;
		iter_best = iter;
  		for (i = 0; i < numcus + 1; i++){
  		  best_sequence[i] = sequence[i];
  		}
 	}

	
}

/************************************************************************
 *                                                                      *
 *                         BACK TO BEST SOLUTION                        *
 *                                                                      *
 ************************************************************************/
	
void back_to_best()

{
	int i;
  
  	for (i = 0; i < numcus + 1; i++){
		sequence[i] = best_sequence[i];
  	}
  

	
}


/************************************************************************
 *                                                                      *
 *               EVALUATE SOLUTION FOR REVERSE CHAIN  		        *
 *                                                                      *
 ************************************************************************/
	
double solution2()

{
	int i;
	double test;

	test = 0;
	for (i = 0; i <= numcus; i++){
		test = test + distance[test_sequence[i]][test_sequence[i+1]];
	}
  
  
  	return test;
	
}


/************************************************************************
 *                                                                      *
 *                           BEST MOVE                                  *
 *                                                                      *
 ************************************************************************/

void best_move(int tabu_size)
{

   
	struct move_info move[((numcus*(numcus-1))/2)];     /* information related to best move      */
     
   
	int i, j, k, best_i, best_j, feasible;
	double best_swap;
   
   
	best_swap = BIG;


	k = -1;
	for (i = 1; i <= numcus-1; i++){
		for (j = i+1; j <= numcus; j++){
			++k;
			move[k].i = i;
			move[k].j = j;
			feasible = feasibility(i, j);
			if (tabu_time[i] > iter || tabu_time[j] > iter || feasible == 0){
				move[k].value = BIG;
			}	
			else {
				move[k].value = evaluate(i,j);
			}
		}
	}
   
   
	for (i = 0; i <= k; i++){
		if (move[i].value < best_swap){
			best_swap = move[i].value;
			best_i = move[i].i;
			best_j = move[i].j;
      		}
	}

 /*  printf("%d %d %lf\n", best_i, best_j, best_swap);*/
   if (best_i>100 || best_j>100) { printf(" Algo paso...\n "); exit(1);}
   execute_move(best_i, best_j, tabu_size);
}



/************************************************************************
 *                                                                      *
 *                           REVERSE CHAIN                             *
 *                                                                      *
 ************************************************************************/

void reverse_chain(int tabu_size)
{

   
	int i, j, k,d, l,best_i, best_j, time_ok, tabu_ok, feasible;
 	int loop_sequence[numcus+2];


  
	double best_swap, loop_solution, test_solution;

   
   
	best_swap = BIG;
	loop_solution = BIG; 
	best_i = 0;
	best_j = 0;
 
	for (i = 1; i <= numcus-2; i++){
		for (j = i+2; j <= numcus; j++){
			for (k = 0; k < j; k++){ 
				test_sequence[k] = sequence[k];
		   	}
		   	d = j-1; 					/*j-1*/
		   	for (k = i; k < j; k++){			/*k < j*/
		   	test_sequence[k] = sequence[d];
		   	--d;
			}
   	
   
			for (k = j; k <= numcus + 1; k++){		/*k = j*/
				test_sequence[k] = sequence[k];
			}
			   
			   
			time_ok = feas();
		  
			if (time_ok == 1){
				tabu_ok = 0;
			   	if (tabu_time[i] <= iter && tabu_time[j] <= iter){
					tabu_ok = 1;
				}
				test_solution = solution2();
				if (tabu_ok == 1 || test_solution < best_solution){
					if (test_solution < loop_solution){
						loop_solution = test_solution;
				   		best_i = i;
				   		best_j = j;
				   		for (k = 0; k <= numcus + 1; k++){
				   	 		loop_sequence[k] = test_sequence[k];
				   		}
				   	}
				}
				
			}
		   
		}
	}
	
	tabu_time[best_i] = iter + tabu_size;
	tabu_time[best_j] = iter + tabu_size;
	
	printf("best_i = %d, best_j = %d\n", best_i, best_j);

	for (i = 0; i <= numcus + 1; i++){
		sequence[i] = loop_sequence[i];
	}
 	iter = iter + 1;


 /*  printf("%d %d %lf\n", best_i, best_j, best_swap);*/
  /* if (best_i>100 || best_j>100) { printf(" Algo paso...\n "); exit(1);}*/
   /*execute_move(best_i, best_j, tabu_size);*/
}

/************************************************************************
 *                                                                      *
 *                       FEASIBILITY CHECK FOR REVERSE CHAIN           *
 *                                                                      *
 ************************************************************************/

int feas()

{
	int feasible, i;
 
	feasible = 1;
		   
	for (i = 1; i <= numcus; i++){
		if ((customers[test_sequence[i]].open > i) || (customers[test_sequence[i]].close < i)){
			feasible = 0;
		}
	}
	return feasible;
}





/************************************************************************
 *                                                                      *
 *                       MOVE EVALUATION                                *
 *                                                                      *
 ************************************************************************/

double evaluate(int i, int j)

{
	int k, temp;
	double d;

	d = 0;
		
	if ((i == (j + 1)) || (j == (i + 1))){
	d = d - distance[sequence[i-1]][sequence[i]];
	d = d - distance[sequence[j]][sequence[j+1]];
	d = d + distance[sequence[i-1]][sequence[j]]; 
	d = d + distance[sequence[i]][sequence[j+1]];
	
	}
	else {
	d = d - distance[sequence[i-1]][sequence[i]];
	d = d - distance[sequence[i]][sequence[i+1]];
	d = d - distance[sequence[j-1]][sequence[j]];
	d = d - distance[sequence[j]][sequence[j+1]];
	d = d + distance[sequence[i-1]][sequence[j]];
	d = d + distance[sequence[j]][sequence[i+1]];
	d = d + distance[sequence[j-1]][sequence[i]];
	d = d + distance[sequence[i]][sequence[j+1]];
	}

	return d;
    
}

/************************************************************************
 *                                                                      *
 *                        EXECUTE MOVE                                  *
 *                                                                      *
 ************************************************************************/

void execute_move(int best_i, int best_j, int tabu_size)

{
	int i, temp;
	
	
	temp = sequence[best_j];
	sequence[best_j] = sequence[best_i];
	sequence[best_i] = temp;
	
	
	tabu_time[best_i] = iter + tabu_size;
	tabu_time[best_j] = iter + tabu_size;
	
	/*for (i = 0; i <= numcus + 1; i++){
	   printf("%d ", sequence[i]);
	}*/
	
	iter++;
}

/************************************************************************
 *                                                                      *
 *                        CHECK FEASIBILITY                             *
 *                                                                      *
 ************************************************************************/

int feasibility(int i, int j)

{
	int f;
	
		
	if (customers[sequence[j]].open > i || customers[sequence[j]].close < i ||
            customers[sequence[i]].open > j || customers[sequence[i]].close < j) {
		f = 0;
	}
	
	else {
		f = 1;
	}
	
	return f;
        
}
/************************************************************************
 *                                                                      *
 *                        ERROR REPORT		                        *
 *                                                                      *
 ************************************************************************/
void nrerror(char *error_text)

{
	fprintf(stderr,"\n%s\n",error_text);
	fprintf(stderr,"... now exiting to system ...\n");
	exit(1);
}

